// FDMVisitor.hpp
//
// Simple class to hold finite difference schemes. In
// this version we work with doubles for convenience.
//
// (C) Datasim Education BV 2007-2011
//


#ifndef FDMVisitor_HPP
#define FDMVisitor_HPP

#include "Sde.hpp"
#include "SdeVisitor.hpp"
#include "Range.cpp"

#include <boost/numeric/ublas/matrix.hpp>		// The matrix class.
#include <boost/numeric/ublas/io.hpp>			// Sending to IO stream.

#include "NormalGenerator.cpp"

template <typename X, typename Time, typename RT, typename Generator>
			class FdmVisitor : public SdeVisitor<X, Time, RT>
{

public: // For convenience
	
	
	// Mesh data
	boost::numeric::ublas::vector<Time> x;

	Time k;		// Time step
	Time T;

	Time sqrk;	// Square root of k (for dW stuff)
	X VOld, VNew;
	Time time;

	// Result path
	boost::numeric::ublas::vector<X> res;

	// Random numbers 
	BoostNormal<Generator> generator;

	// Number of steps
	long N;

	Sde<X,Time,RT> sde;

public:
	FdmVisitor(long NSteps, const Sde<X,Time,RT>& mySde, const Generator& generator);

	virtual boost::numeric::ublas::vector<X>& path();

};

template <typename X, typename Time, typename RT,  typename Generator>
	class ExplicitEuler : public FdmVisitor<X,Time,RT, Generator>
{ // Explicit Euler method

private:
	
	
public:
	ExplicitEuler(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class ExplicitEulerTypeII : public FdmVisitor<X,Time,RT, Generator>
{ // Explicit Euler by creating a random array beforehand

private:
	
	boost::numeric::ublas::vector<Time> dW2; 

public:
	ExplicitEulerTypeII(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class ExplicitEulerMM : public FdmVisitor<X,Time,RT, Generator>
{ // Explicit Euler by Momen matching (quadratic resampling

private:
	
	boost::numeric::ublas::vector<Time> dW2; 

public:
	ExplicitEulerMM(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class PredictorCorrector : public FdmVisitor<X,Time,RT, Generator>
{ // Adapted Predictor Corrector method

private:
	
	X A, B;
	X VMid;	// Predictor value

public:
	PredictorCorrector(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator,
						X alpha, X beta);

	void Visit(Sde<X,Time,RT>& sde);
};
			
template <typename X, typename Time, typename RT,  typename Generator>
	class PredictorCorrectorClassico : public FdmVisitor<X,Time,RT, Generator>
{ // Standard Predictor Corrector method: 1) no drift adjust, 2) trapezoidal averaging

private:
	
	X A, B;
	X VMid;	// Predictor value

public:
	PredictorCorrectorClassico(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator,
						X alpha, X beta);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class RichardsonEuler : public FdmVisitor<X,Time,RT, Generator>
{ // Richardson extrapolation

private:
	
		// Values in one-step method
		X VOld, VNew;
		
		boost::numeric::ublas::vector<Time> res2;
		boost::numeric::ublas::vector<Time> x2;
		boost::numeric::ublas::vector<Time> dW2; 

		double k2, sqrk2;

public:
	RichardsonEuler(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class Milstein : public FdmVisitor<X,Time,RT, Generator>
{ // Richardson extrapolation

private:
	
	
public:
	Milstein(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator);

	void Visit(Sde<X,Time,RT>& sde);
};

template <typename X, typename Time, typename RT,  typename Generator>
	class KarhunenLoeve : public FdmVisitor<X,Time,RT, Generator>
{ // Richardson extrapolation

private:
	
	double tol, sqrT;
	long Ntrunc; // Computed number of needed terms
	boost::numeric::ublas::vector<Time> dW2; 

	X orthogonalFunction(Time t, long n);
	X KLExpansion(Time s, Time t);

public:
	class KarhunenLoeve(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator, double tolerance);

	void Visit(Sde<X,Time,RT>& sde);
};

// Predictor-Corrector with Karhunen-Loeve
template <typename X, typename Time, typename RT,  typename Generator>
	class PredictorCorrectorKL : public FdmVisitor<X,Time,RT, Generator>
{ // Standard Predictor Corrector method

private:
	
	X A, B;
	X VMid;	// Predictor value

	double tol, sqrT;
	long Ntrunc; // Computed number of needed terms
	boost::numeric::ublas::vector<Time> dW2; 

	X orthogonalFunction(Time t, long n);
	X KLExpansion(Time s, Time t);

public:
	PredictorCorrectorKL(long NSteps, Sde<X,Time,RT>& sde, const Generator& generator,
						X alpha, X beta, double tolerance);

	void Visit(Sde<X,Time,RT>& sde);
};

#endif
