// MCTypeDMediator.hpp
//
// Control class for the one-factor SDE and its
// approximation in FDM.
//
// Clients (e.g. a main program) use this class as a 
// Facade because it has a well-defined interface.
// The mediatoir uses a factory to initialise all relevant
// variables.
//
// 2007-3-10 DD Kick-off and initial version
// 2007-7-24 DD Debugging
// 2011-12-5 DD payoff --> signals
// 2011-12-9 DD use uBLAS vector
// 2011-12-11 DD generic RMG class from Boost
//
// This class plays the role of the Director in the Builder
// pattern (if we decide to use it).
//
// (C) Datasim Education BV 2007-2011
//

#ifndef MCTypeDMediator_HPP
#define MCTypeDMediator_HPP

#include <boost/random.hpp>

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>			// Sending to I/O stream.

#include "MCPostProcess.hpp"
#include "Sde.hpp"
#include "FdmVisitor.hpp"

#include <boost/function.hpp>
#include <boost/signals.hpp>


// Namespaces typedefs etc.
namespace ublas=boost::numeric::ublas;
typedef boost::function<double (double S)> Payoff;	// The interface for payoff function
typedef boost::signal<void (long index)> Progress;	// The progress bar (can be ON/OFF)

template <typename Generator>
				class MCTypeDMediator
{
private:
		long NSim;				// Number of simulations, needed for discounting

		Payoff payoff;
		
		Progress prog;
		FdmVisitor<double,double,double,Generator>* fdm;

public:
		MCTypeDMediator(FdmVisitor<double,double,double, Generator>& myFdm, 
			long NSimulations, const Payoff& optionPayoff, const boost::function<void (long val)>& progress, bool show)
			 : payoff(optionPayoff), prog(Progress()) 
		{
			NSim = NSimulations;
			if (show == true)
			{
				prog.connect(progress);
			}
			
			fdm = &myFdm;			
		}

		void price()
		{
			// A. Loop over each iteration
			// B. Find value at t = T
			// C. Return the vectors and calc. payoff vectors + average
			// D. Discount the value

			
			// Create the random numbers
			long N = fdm->N;
		
			ublas::vector<double> TerminalValue(NSim, 0.0); // Array of values at t = T
	
			// A.
			for (long i = 0; i < NSim; ++i)
			{ // Calculate a path at each iteration
		
				prog(i);
			
				// Compute the current path and get value at t = T
				// For more complicated payoffs we have to send the complete path.
				TerminalValue[i] = payoff(fdm->path()[fdm->path().size()-1]);
			}
	
			// Send statistics Display information
			MCReporter mcr;

			// V2: signals2
			boost::signal<void (Status)> slotControl;
			boost::signal<void (const boost::numeric::ublas::vector<double>& arr)> slotData;

			// Connect signals to slots. N.B. use Boost references, otherwise a copy is
			// made and you will get incorrect results.
			slotControl.connect(boost::ref(mcr)); // Create a reference to mcr
			slotData.connect(boost::ref(mcr));

			//  C. Take the average; price will be in the slot
			slotControl(START);						// Signal to start process, timer
				slotData(TerminalValue);			// Marshall computed data to postprocessor	
			slotControl(STOP);						// Signal to stop receiving data
				
		}
};

#endif
