// MCPostProcess.hpp
//
// A Slot class that gets data from a MC signal 'emitter'. This corresponds to a MIS
// system that postprocesses data from the MC engine.
//
// (C) Datasim Education BV 2011
//

#ifndef MCPost_hpp
#define MCPost_hpp

#include <iostream>
#include <cmath>
#include <algorithm>
#include <numeric>
#include <list>
#include <vector>
#include <boost/signals.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_io.hpp>
#include <boost/timer.hpp>
#include "SdeOneFactor.hpp"
enum Status {START, STOP};


// Some statistics-based functions
template <typename V>
	boost::tuple<double, double> SumSquaresAndSum(const V& container)
{ // Sum and sum of squares of a container

	double sum = 0.0;
	double sumSquares = 0.0;

	std::for_each(container.begin(), container.end(), [&sum, &sumSquares] (double d) { sum += d; sumSquares += d*d;});

	return boost::tuple<double, double> (sum, sumSquares);
}


template <typename V>
	boost::tuple<double, double> StandardDeviationAndError(const V& container, double r, double T)
{ // Standard deviation and error

	double sum;;
	double sumSquares;
	double M = double(container.size());

	boost::tie(sum, sumSquares) = SumSquaresAndSum<V>(container);

	// Now compute SD and SE
	double SD = (sqrt(sumSquares - (sum*sum)/M) * exp(-2.0*r*T)) / (M - 1.0);
	double SE = SD/sqrt(M);

	return boost::tuple<double, double> (SD, SE);
}
	
struct MCMisAgent
{
	boost::timer t;

	MCMisAgent() : t(boost::timer()) {}

	void operator () (Status s)
	{
		if (s == START)
		{
			cout << "\n\nMIS Configuration information for the MC Engine, starting up:\n ";
		}
		else if (s == STOP)
		{
			cout << "\nMIS, final results: ";
			cout<<"Time for MC simulation: " << t.elapsed() << endl;
		}
	}

	void operator () (long& NSIM, long& NT, bool& progressBar)
	{ // Simulation parameters

		cout << "NT: ";  cin >> NT;
		cout << "NSIM: "; cin >> NSIM;
		cout << "Progress Yes/N: "; char a; cin >> a;
		if (a == 'y' || a == 'Y')
		{
			progressBar = true;
		}
		else
		{
			progressBar = false;
		}

	}

};


struct MCReporter
{
	
	boost::numeric::ublas::vector<double> arr;

	MCReporter() : arr(boost::numeric::ublas::vector<double>())
	{

		
	}


	void operator () (Status s)
	{
		if (s == START)
		{
			std::cout << "Start " << arr.size() << std::endl;
		}
		else if (s == STOP)
		{
			// Print using lambda functions
			using namespace OneFactorSDE;
		
			// Calculate the price
			//	double initVal = 0.0;
			//	double sum = std::accumulate(arr.begin(), arr.end(), initVal);
			
			// Sum of option values at terminal time T
			double sum = 0.0;
			std::for_each(arr.begin(), arr.end(), [&sum] (double d) { sum += d; });

			double price = exp(-r*T) * sum / double (arr.size());
			std::cout << "\nPrice: " << price << ". ";

			boost::tuple<double, double> answer = StandardDeviationAndError<boost::numeric::ublas::vector<double>> (arr, r, T);
			std::cout << " StdDeviation, StdError " << answer << std::endl;

		}
	}

	void operator () (const boost::numeric::ublas::vector<double>& packetArr)
	{
		arr = packetArr;
	}
};


#endif