// TestMC.cpp
//
// Some FD schemes for MC simulation and 1-factor option pricing.
//
// The design is based on Duffy's Domain Architectures (in this case RAT). Module
// interface by means of Boot signals and function.
//
// Maon subsysystems:
//
//		MC engine (main())
//	    Payoff
//      FDM module (with fd schemes, SDE and RNG)
//		Display
//		MIS Agent
// 
// For n-factor models we need to define the signature of signals and slots (mostly data) by making it more 
// generic.
//
// (C) Datasim Education BV 2006-2011

#include "Mesher.hpp"
using namespace std;

#include <cmath>

#include "FdmVisitor.cpp"
#include "MCMediator.hpp"
#include "SdeOneFactor.hpp"
#include <boost/bind.hpp>
#include <boost/function.hpp>

void ProgressPrint(long index)
{ // 'Subsystem' that implements progress and run-time statistics

	if ((index/1000)*1000 == index)
	{
		std::cout << index << ", ";
	}
}


int main()
{
	// ACTIVATE THE MDODEL OF CHOICE HERE!

	using namespace OneFactorSDE;
	//using namespace CIRSDE;

	// Create the basic SDE (Context class)
	Range<double> range (0.0, T);
	Sde<double,double,double> sde(initialCondition, range, drift, driftCorrected, diffusion, diffusionDerivative);
	

	// Management module for settings etc.
	MCMisAgent misAgent;
	
	// Settings and Factories from MIS
	long NSimulations = 10000; long N = 200; bool progressDisplay;
	misAgent(NSimulations, N, progressDisplay);
	

	// Factories (compile-time because we are using templates)
/*	boost::lagged_fibonacci607 rng;
	ExplicitEuler<double,double,double,boost::lagged_fibonacci607> fdm(N, sde, rng); // N == number of intervals
	MCTypeDMediator<boost::lagged_fibonacci607> mediator(fdm, NSimulations, myPayOffFunction, ProgressPrint, progressDisplay);

*/

	boost::mt19937 rng;
	rng.seed(static_cast<boost::uint32_t> (std::time(0)));

	// The schemes; N == number of intervals
	//..
	//ExplicitEuler<double,double,double,boost::mt19937> fdm(N, sde, rng); // Good ole one
	//ExplicitEulerTypeII<double,double,double,boost::mt19937> fdm(N, sde, rng); // Precompute rn array
	//ExplicitEulerMM<double,double,double,boost::mt19937> fdm(N, sde, rng); // MM == Method of moments
	RichardsonEuler<double,double,double,boost::mt19937> fdm(N, sde, rng); // Higher order accuracy
	//PredictorCorrector<double,double,double,boost::mt19937> fdm(N, sde, rng, 0.5, 0.5); // PC with adjusted drift
	//PredictorCorrectorClassico<double,double,double,boost::mt19937> fdm(N, sde, rng, 0.5, 0.5); // No drift adjustmen
	//Milstein<double,double,double,boost::mt19937> fdm(N, sde, rng); // Slight improvement on Euler

	// Karhunen-Loeve family
	double tol = 0.01; // Determines number of terms in truncated KL expansion
	//cout << "tol: "; cin >> tol;
	//KarhunenLoeve<double,double,double,boost::mt19937> fdm(N, sde, rng, tol); // Euler + KL
	//KarhunenLoeve<double,double,double,boost::lagged_fibonacci607> fdm(N, sde, rng, tol); // Euler + KL
	//PredictorCorrectorKL<double,double,double,boost::mt19937> fdm(N, sde, rng, 0.5, 0.5, tol); // PC + KL

	MCTypeDMediator<boost::mt19937> mediator(fdm, NSimulations, myPayOffFunction, ProgressPrint, progressDisplay);
			
	boost::signal<void (Status)> slotControl;
		
	try
	{
		// Connect signals to slots. N.B. use Boost references, otherwise a copy is
		// made and you will get incorrect results.
		slotControl.connect(boost::ref(misAgent)); // Create a reference to mcr
		
		slotControl(START);						// Signal to start process, timer

		mediator.price();						// GET THE PRICE, SD, SE

		slotControl(STOP);						// Signal to stop receiving data

	}
	catch(std::exception& exception)
	{ 
		cout << exception.what() << endl;
		exit(1);
	}

	return 0;
}